"""Code and data structures for managing source directives."""

import bisect
import collections
import dataclasses
import logging
import sys
from typing import AbstractSet, Optional

from pytype import blocks

# directors.parser uses the stdlib ast library, which is much faster than
# libcst, but we rely on ast features that are new in Python 3.9.
# pylint: disable=g-import-not-at-top
if sys.version_info[:2] >= (3, 9):
  from pytype.directors import parser
else:
  from pytype.directors import parser_libcst as parser
# pylint: enable=g-import-not-at-top

log = logging.getLogger(__name__)

SkipFileError = parser.SkipFileError
parse_src = parser.parse_src

_ALL_ERRORS = "*"  # Wildcard for disabling all errors.

_FUNCTION_CALL_ERRORS = frozenset((
    # A function call may implicitly access a magic method attribute.
    "attribute-error",
    "duplicate-keyword",
    # Subscripting an annotation is a __getitem__ call.
    "invalid-annotation",
    "missing-parameter",
    "not-instantiable",
    "wrong-arg-count",
    "wrong-arg-types",
    "wrong-keyword-args",
    "unsupported-operands",
))

_ALL_ADJUSTABLE_ERRORS = _FUNCTION_CALL_ERRORS.union((
    "annotation-type-mismatch",
    "bad-return-type",
    "bad-yield-annotation",
    "container-type-mismatch",
    "not-supported-yet",
))


class _DirectiveError(Exception):
  pass


class _LineSet:
  """A set of line numbers.

  The data structure is optimized to represent the union of a sparse set
  of integers and ranges of non-negative integers.  This supports the two styles
  of directives: those after a statement apply only to that line and those on
  their own line apply until countered by the opposing directive.
  """

  def __init__(self):
    # Map of line->bool for specific lines, takes precedence over _transitions.
    self._lines = {}
    # A sorted list of the lines at which the range state changes
    # polarity.  It is assumed to initially be false (not in a range).
    # Even positions represent the start of a range, odd positions represent
    # the end of a range.  Thus [2, 5, 10, 12] would include lines 2, 3, 4, 10,
    # and 11.  If the length is odd, then an end of maxint is implied, thus
    # [2, 5, 10] would disable lines 2, 3, 4, 10, 11, 12, ...
    self._transitions = []

  @property
  def lines(self):
    return self._lines

  def set_line(self, line, membership):
    """Set whether a given line is a member of the set."""
    self._lines[line] = membership

  def start_range(self, line, membership):
    """Start a range of lines that are either included/excluded from the set.

    Args:
      line: A line number.
      membership: If True, lines >= line are included in the set (starting
        a range), otherwise they are excluded (ending a range).

    Raises:
      ValueError: if line is less than that of a previous call to start_range().
    """
    last = self._transitions[-1] if self._transitions else -1
    # Assert that lines are monotonically increasing.  This simplifies the
    # logic of adding new lines and ensures that _ranges is sorted.
    if line < last:
      raise ValueError("Line number less than previous start_range() call.")
    # Determine previous membership state (True if the last range has an
    # indefinite end).
    previous = (len(self._transitions) % 2) == 1
    if membership == previous:
      return  # Redundant with previous state, do nothing.
    elif line == last:
      # We have either enable/disable or disable/enable on the same line,
      # cancel them out by popping the previous transition.
      self._transitions.pop()
    else:
      # Normal case - add a transition at this line.
      self._transitions.append(line)

  def __contains__(self, line):
    """Return if a line is a member of the set."""
    # First check for an entry in _lines.
    specific = self._lines.get(line)
    if specific is not None:
      return specific
    # Find the position in _ranges for line.  The polarity of this position
    # determines whether we are inside a range (odd) or outside (even).
    pos = bisect.bisect(self._transitions, line)
    return (pos % 2) == 1

  def get_disable_after(self, line):
    """Get an unclosed disable, if any, that starts after line."""
    if len(self._transitions) % 2 == 1 and self._transitions[-1] >= line:
      return self._transitions[-1]
    return None


class _BlockRanges:
  """A collection of possibly nested start..end ranges from AST nodes."""

  def __init__(self, start_to_end_mapping):
    self._starts = sorted(start_to_end_mapping)
    self._start_to_end = start_to_end_mapping
    self._end_to_start = {v: k for k, v in start_to_end_mapping.items()}

  def has_start(self, line):
    return line in self._start_to_end

  def has_end(self, line):
    return line in self._end_to_start

  def find_outermost(self, line):
    """Find the outermost interval containing line."""
    i = bisect.bisect_left(self._starts, line)
    num_intervals = len(self._starts)
    if i:
      if i < num_intervals and self._starts[i] == line:
        # line number is start of interval.
        start = self._starts[i]
      else:
        # Skip nested intervals
        while (
            1 < i <= num_intervals and
            self._start_to_end[self._starts[i - 1]] < line):
          i -= 1
        start = self._starts[i - 1]
      return start, self._start_to_end[start]
    return None, None

  def adjust_end(self, old_end, new_end):
    start = self._end_to_start[old_end]
    self._start_to_end[start] = new_end
    del self._end_to_start[old_end]
    self._end_to_start[new_end] = start


def _collect_bytecode(ordered_code):
  bytecode_blocks = []
  stack = [ordered_code]
  while stack:
    code = stack.pop()
    bytecode_blocks.append(code.original_co_code)
    for const in code.co_consts:
      if isinstance(const, blocks.OrderedCode):
        stack.append(const)
  return bytecode_blocks


def _adjust_line_number(line, allowed_lines, min_line):
  adjusted_line = line
  while adjusted_line not in allowed_lines and adjusted_line >= min_line:
    adjusted_line -= 1
  return adjusted_line if adjusted_line >= min_line else None


def _is_function_call(opcode_name):
  return opcode_name.startswith("CALL_") or opcode_name in {
      "BINARY_SUBSCR",
      "COMPARE_OP",
      "FOR_ITER",
  }


def _is_load_attribute_op(opcode_name):
  """Checks whether the opcode loads an attribute."""
  return (opcode_name.startswith("GET_") or
          opcode_name.startswith("UNPACK_") or
          opcode_name in {
              "LOAD_ATTR",
              "LOAD_METHOD",
              "SETUP_WITH",
          })


def _is_return_op(opcode_name):
  return opcode_name.startswith("YIELD_") or opcode_name == "RETURN_VALUE"


@dataclasses.dataclass
class _OpcodeLines:
  """Stores opcode line numbers for Director.adjust_line_numbers()."""

  store_lines: AbstractSet[int]
  make_function_lines: AbstractSet[int]
  load_attr_lines: AbstractSet[int]
  return_lines: AbstractSet[int]
  call_lines: AbstractSet[int]

  @classmethod
  def from_code(cls, code):
    """Builds an _OpcodeLines from a code object."""
    store_lines = set()
    make_function_lines = set()
    load_attr_lines = set()
    return_lines = set()
    call_lines = set()
    for block in _collect_bytecode(code):
      for opcode in block:
        if opcode.name.startswith("STORE_"):
          store_lines.add(opcode.line)
        elif opcode.name == "MAKE_FUNCTION":
          make_function_lines.add(opcode.line)
        elif _is_load_attribute_op(opcode.name):
          load_attr_lines.add(opcode.line)
        elif _is_return_op(opcode.name):
          return_lines.add(opcode.line)
        elif _is_function_call(opcode.name):
          call_lines.add(opcode.line)
    return cls(store_lines, make_function_lines, load_attr_lines, return_lines,
               call_lines)


class Director:
  """Holds all of the directive information for a source file."""

  def __init__(self, src_tree, errorlog, filename, disable, code):
    """Create a Director for a source file.

    Args:
      src_tree: The source text as an ast.
      errorlog: An ErrorLog object.  Directive errors will be logged to the
        errorlog.
      filename: The name of the source file.
      disable: List of error messages to always ignore.
      code: Optionally, bytecode for adjusting line numbers. If provided,
        directives will be moved to lines at which corresponding opcodes are
        present. Otherwise, directives will be moved to the starting line of
        their containing statement.
    """
    self._filename = filename
    self._errorlog = errorlog
    self._type_comments = {}  # Map from line number to type comment.
    self._variable_annotations = {}  # Map from line number to annotation.
    # Lines that have "type: ignore".  These will disable all errors, and in
    # the future may have other impact (such as not attempting an import).
    self._ignore = _LineSet()
    # Map from error name to lines for which that error is disabled.  Note
    # that _ALL_ERRORS is essentially a wildcard name (it matches all names).
    self._disables = collections.defaultdict(_LineSet)
    # Line numbers of decorators. Since this is used to mark a class or function
    # as decorated, stacked decorators will record the one closest to the
    # definition (i.e. the last one). The python bytecode uses this line number
    # for all the stacked decorator invocations.
    self._decorators = set()
    # Apply global disable, from the command line arguments:
    for error_name in disable:
      self._disables[error_name].start_range(0, True)
    # Store function ranges and return lines to distinguish explicit and
    # implicit returns (the bytecode has a `RETURN None` for implcit returns).
    self._return_lines = set()
    self._function_ranges = _BlockRanges({})
    # Parse the source code for directives.
    self._parse_src_tree(src_tree, code)

  @property
  def type_comments(self):
    return self._type_comments

  @property
  def annotations(self):
    # It's okay to overwrite type comments with variable annotations here
    # because _FindIgnoredTypeComments in vm.py will flag ignored comments.
    return {**self._type_comments, **self._variable_annotations}

  @property
  def ignore(self):
    return self._ignore

  @property
  def decorators(self):
    return self._decorators

  def _parse_src_tree(self, src_tree, code):
    """Parse a source file, extracting directives from comments."""
    visitor = parser.visit_src_tree(src_tree)
    # TODO(rechen): This check can be removed once parser_libcst is gone.
    if not visitor:
      return
    if code:
      opcode_lines = _OpcodeLines.from_code(code)
    else:
      opcode_lines = None

    self._return_lines = visitor.returns
    self._function_ranges = _BlockRanges(visitor.function_ranges)

    for line_range, group in visitor.structured_comment_groups.items():
      for comment in group:
        if comment.tool == "type":
          self._process_type(comment.line, comment.data, comment.open_ended,
                             line_range, opcode_lines)
        else:
          assert comment.tool == "pytype"
          try:
            self._process_pytype(comment.line, comment.data, comment.open_ended,
                                 line_range, opcode_lines)
          except _DirectiveError as e:
            self._errorlog.invalid_directive(
                self._filename, comment.line, str(e))
        # Make sure the function range ends at the last "interesting" line.
        if (not isinstance(line_range, parser.Call) and
            self._function_ranges.has_end(line_range.end_line)):
          if opcode_lines:
            end = _adjust_line_number(
                line_range.end_line, opcode_lines.return_lines,
                line_range.start_line)
          else:
            end = line_range.start_line
          self._function_ranges.adjust_end(line_range.end_line, end)

    for annot in visitor.variable_annotations:
      if opcode_lines:
        final_line = _adjust_line_number(
            annot.end_line, opcode_lines.store_lines, annot.start_line)
        if not final_line:
          log.error("No STORE_* opcode found for annotation %r on line %d",
                    annot, annot.end_line)
          continue
      else:
        final_line = annot.start_line
      self._variable_annotations[final_line] = annot.annotation

    for decorator in visitor.decorators:
      # The MAKE_FUNCTION opcode is usually at the 'def' line but pre-3.8 was
      # sometimes somewhere in the last decorator's line range.
      if opcode_lines:
        final_line = _adjust_line_number(
            decorator.end_line, opcode_lines.make_function_lines,
            decorator.start_line)
        if not final_line:
          log.warning("No MAKE_FUNCTION opcode found for decorator on line %d",
                      decorator.end_line)
          continue
      else:
        final_line = decorator.end_line
      self._decorators.add(final_line)

    if visitor.defs_start is not None:
      disables = list(self._disables.items())
      # Add "# type: ignore" to the list of disables that we check.
      disables.append(("Type checking", self._ignore))
      for name, lineset in disables:
        lineno = lineset.get_disable_after(visitor.defs_start)
        if lineno is not None:
          self._errorlog.late_directive(self._filename, lineno, name)

  def _process_type(
      self, line: int, data: str, open_ended: bool,
      line_range: parser.LineRange, opcode_lines: Optional[_OpcodeLines]):
    """Process a type: comment."""
    is_ignore = parser.IGNORE_RE.match(data)
    if not is_ignore and line != line_range.end_line:
      # Warn and discard type comments placed in the middle of expressions.
      self._errorlog.ignored_type_comment(self._filename, line, data)
      return
    if opcode_lines:
      final_line = _adjust_line_number(
          line, opcode_lines.store_lines, line_range.start_line) or line
    else:
      final_line = line_range.start_line
    if is_ignore:
      if open_ended:
        self._ignore.start_range(line, True)
      else:
        self._ignore.set_line(line, True)
        self._ignore.set_line(final_line, True)
    else:
      if final_line in self._type_comments:
        # If we have multiple type comments on the same line, take the last one,
        # but add an error to the log.
        self._errorlog.invalid_directive(
            self._filename, line, "Multiple type comments on the same line.")
      self._type_comments[final_line] = data

  def _process_pytype(
      self, line: int, data: str, open_ended: bool,
      line_range: parser.LineRange, opcode_lines: Optional[_OpcodeLines]):
    """Process a pytype: comment."""
    if not data:
      raise _DirectiveError("Invalid directive syntax.")
    for option in data.split():
      # Parse the command.
      try:
        command, values = option.split("=", 1)
        values = values.split(",")
      except ValueError as e:
        raise _DirectiveError("Invalid directive syntax.") from e
      # Additional commands may be added in the future.  For now, only
      # "disable" and "enable" are supported.
      if command == "disable":
        disable = True
      elif command == "enable":
        disable = False
      else:
        raise _DirectiveError(f"Unknown pytype directive: '{command}'")
      if not values:
        raise _DirectiveError(
            "Disable/enable must specify one or more error names.")

      def keep(error_name):
        if isinstance(line_range, parser.Call):
          return error_name in _FUNCTION_CALL_ERRORS
        else:
          return True

      for error_name in values:
        if (error_name == _ALL_ERRORS or
            self._errorlog.is_valid_error_name(error_name)):
          if not keep(error_name):
            # Skip the directive if we are in a line range that is irrelevant to
            # it. (Every directive is also recorded in a base LineRange that is
            # never skipped.)
            continue
          lines = self._disables[error_name]
          if open_ended:
            lines.start_range(line, disable)
          else:
            final_line = self._adjust_line_number_for_pytype_directive(
                line, error_name, line_range, opcode_lines)
            if final_line != line:
              # Set the disable on the original line so that, even if we mess up
              # adjusting the line number, silencing an error by adding a
              # disable to the exact line the error is reported on always works.
              lines.set_line(line, disable)
            lines.set_line(final_line, disable)
        else:
          self._errorlog.invalid_directive(
              self._filename, line, f"Invalid error name: '{error_name}'")

  def _adjust_line_number_for_pytype_directive(
      self, line: int, error_class: str, line_range: parser.LineRange,
      opcode_lines: Optional[_OpcodeLines]):
    """Adjusts the line number for a pytype directive."""
    if error_class not in _ALL_ADJUSTABLE_ERRORS:
      return line
    if not opcode_lines:
      return line_range.start_line
    if error_class == "annotation-type-mismatch":
      allowed_lines = (
          opcode_lines.store_lines | opcode_lines.make_function_lines)
    elif error_class == "attribute-error":
      allowed_lines = opcode_lines.load_attr_lines
    elif error_class == "bad-return-type":
      allowed_lines = opcode_lines.return_lines
    elif error_class == "bad-yield-annotation":
      allowed_lines = opcode_lines.make_function_lines
    elif error_class == "invalid-annotation":
      allowed_lines = opcode_lines.make_function_lines
    elif error_class == "not-supported-yet":
      allowed_lines = opcode_lines.store_lines
    elif error_class == "unsupported-operands":
      allowed_lines = opcode_lines.store_lines | opcode_lines.call_lines
    else:
      allowed_lines = opcode_lines.call_lines
    return _adjust_line_number(
        line, allowed_lines, line_range.start_line) or line

  def filter_error(self, error):
    """Return whether the error should be logged.

    This method is suitable for use as an error filter.

    Args:
      error: An error._Error object.

    Returns:
      True iff the error should be included in the log.
    """
    # Always report errors that aren't for this file or do not have a line
    # number.
    if error.filename != self._filename or error.lineno is None:
      return True
    if (error.name == "bad-return-type" and
        error.opcode_name == "RETURN_VALUE" and
        error.lineno not in self._return_lines):
      # We have an implicit "return None". Adjust the line number to the last
      # line of the function.
      _, end = self._function_ranges.find_outermost(error.lineno)
      if end:
        error.set_lineno(end)
    # Treat line=0 as below the file, so we can filter it.
    line = error.lineno or sys.maxsize
    # Report the error if it isn't subject to any ignore or disable.
    return (line not in self._ignore and
            line not in self._disables[_ALL_ERRORS] and
            line not in self._disables[error.name])



from typing import Any

from .elements import conv as conv

class ConventionDict:
    const: Any
    table: Any
    convention: Any
    def __init__(self, const, table, convention) -> None: ...
    def __getitem__(self, key): ...



from typing import AnyStr, Iterable

_EitherStr = str | unicode

def fnmatch(filename: _EitherStr, pattern: _EitherStr) -> bool: ...
def fnmatchcase(filename: _EitherStr, pattern: _EitherStr) -> bool: ...
def filter(names: Iterable[AnyStr], pattern: _EitherStr) -> list[AnyStr]: ...
def translate(pattern: AnyStr) -> AnyStr: ...



"""AST visiting."""


class BaseVisitor:
  """A base class for writing AST visitors.

    Subclasses should define {visit,enter,leave}_X to process nodes of type X.
    If a visit method returns a non-None value, the visited node is replaced
    with that value.

    Attributes:
      _ast: Any module whose interface matches the standard ast library, such as
        typed_ast. The same module must be used to generate the AST to visit.
  """

  def __init__(self, ast, visit_decorators=True):
    self._ast = ast
    maybe_decorators = ["decorator_list"] if visit_decorators else []
    self._node_children = {
        self._ast.Module: ["body"],
        self._ast.ClassDef: maybe_decorators + ["bases", "body"],
        self._ast.FunctionDef: maybe_decorators + ["body", "args", "returns"],
        self._ast.Assign: ["targets", "value"],
    }

  def visit(self, node):
    """Does a post-order traversal of the AST."""
    if isinstance(node, self._ast.AST):
      self.enter(node)
      for k, v in self._children(node):
        ret = self.visit(v)
        if ret is not None:
          setattr(node, k, ret)
      out = self._call_visitor(node)
      self.leave(node)
      if out is not None:
        return out
    elif isinstance(node, list):
      for i, v in enumerate(node):
        ret = self.visit(v)
        if ret is not None:
          node[i] = ret

  def _children(self, node):
    """Children to recurse over."""
    ks = self._node_children.get(node.__class__)
    if ks:
      return [(k, getattr(node, k)) for k in ks]
    else:
      return self._ast.iter_fields(node)

  def _call_visitor(self, node):
    method = "visit_" + node.__class__.__name__
    visitor = getattr(self, method, self.generic_visit)
    return visitor(node)

  def generic_visit(self, node):
    """Called when no visit function is found for a node type."""
    del node  # unused

  def enter(self, node):
    """Does a pre-order traversal of the AST."""
    method = "enter_" + node.__class__.__name__
    visitor = getattr(self, method, None)
    if visitor:
      return visitor(node)

  def leave(self, node):
    """Called after visit() to do any cleanup that enter() needs."""
    method = "leave_" + node.__class__.__name__
    visitor = getattr(self, method, None)
    if visitor:
      visitor(node)



"""LibCST-based parser."""

import collections
import dataclasses
import logging
import re

import libcst
from pytype import utils

log = logging.getLogger(__name__)

# Also supports mypy-style ignore[code, ...] syntax, treated as regular ignores.
IGNORE_RE = re.compile(r"^ignore(\[.+\])?$")

_DIRECTIVE_RE = re.compile(r"#\s*(pytype|type)\s*:\s?([^#]*)")


class SkipFileError(Exception):
  """Exception thrown if we encounter "pytype: skip-file" in the source code."""


@dataclasses.dataclass(frozen=True)
class LineRange:
  start_line: int
  end_line: int


@dataclasses.dataclass(frozen=True)
class Call(LineRange):
  """Tag to identify function calls."""


@dataclasses.dataclass(frozen=True)
class _StructuredComment:
  """A structured comment.

  Attributes:
    line: The line number.
    tool: The tool label, e.g., "type" for "# type: int".
    data: The data, e.g., "int" for "# type: int".
    open_ended: True if the comment appears on a line by itself (i.e., it is
     open-ended rather than attached to a line of code).
  """
  line: int
  tool: str
  data: str
  open_ended: bool


@dataclasses.dataclass(frozen=True)
class _VariableAnnotation(LineRange):
  annotation: str


class _ParseVisitor(libcst.CSTVisitor):
  """Visitor for parsing a source tree.

  Attributes:
    structured_comment_groups: Ordered map from a line range to the "type:" and
      "pytype:" comments within the range. Line ranges come in several flavors:
      * Instances of the base LineRange class represent single logical
        statements. These ranges are ascending and non-overlapping and record
        all structured comments found.
      * Instances of the Call subclass represent function calls. These ranges
        are ascending by start_line but may overlap and only record "pytype:"
        comments.
    variable_annotations: Sequence of PEP 526-style variable annotations with
      line numbers.
    decorators: Sequence of lines at which decorated functions are defined.
    defs_start: The line number at which the first class or function definition
      appears, if any.
  """

  METADATA_DEPENDENCIES = (libcst.metadata.PositionProvider,
                           libcst.metadata.ParentNodeProvider,)

  def __init__(self):
    self.structured_comment_groups = collections.OrderedDict()
    self.variable_annotations = []
    self.decorators = []
    self.defs_start = None
    self.returns = set()
    self.function_ranges = {}

  def _get_containing_groups(self, start_line, end_line=None):
    """Get _StructuredComment groups that fully contain the given line range."""
    end_line = end_line or start_line
    # Since the visitor processes the source file roughly from top to bottom,
    # the given line range should be within a recently added comment group. We
    # also keep the groups ordered. So we do a reverse search and stop as soon
    # as we hit a statement that does not overlap with our given range.
    for line_range, group in reversed(self.structured_comment_groups.items()):
      if (line_range.start_line <= start_line and
          end_line <= line_range.end_line):
        yield (line_range, group)
      elif (not isinstance(line_range, Call) and
            line_range.end_line < start_line):
        return

  def _has_containing_group(self, start_line, end_line=None):
    for line_range, _ in self._get_containing_groups(start_line, end_line):
      if not isinstance(line_range, Call):
        return True
    return False

  def _add_structured_comment_group(self, start_line, end_line, cls=LineRange):
    """Adds an empty _StructuredComment group with the given line range."""
    if cls is LineRange and self._has_containing_group(start_line, end_line):
      return
    # We keep structured_comment_groups ordered by inserting the new line range
    # at the end, then absorbing line ranges that the new range contains and
    # calling move_to_end() on ones that should come after it. We encounter line
    # ranges in roughly ascending order, so this reordering is not expensive.
    keys_to_absorb = []
    keys_to_move = []
    for line_range in reversed(self.structured_comment_groups):
      if (cls is LineRange and
          start_line <= line_range.start_line and
          line_range.end_line <= end_line):
        if type(line_range) is LineRange:  # pylint: disable=unidiomatic-typecheck
          keys_to_absorb.append(line_range)
        else:
          keys_to_move.append(line_range)
      elif line_range.start_line > start_line:
        keys_to_move.append(line_range)
      else:
        break
    self.structured_comment_groups[cls(start_line, end_line)] = new_group = []
    for k in reversed(keys_to_absorb):
      new_group.extend(self.structured_comment_groups[k])
      del self.structured_comment_groups[k]
    for k in reversed(keys_to_move):
      self.structured_comment_groups.move_to_end(k)

  def _process_comment(self, line, comment, open_ended):
    """Process a single comment."""
    matches = list(_DIRECTIVE_RE.finditer(comment))
    if not matches:
      return
    is_nested = matches[0].start(0) > 0
    for m in matches:
      tool, data = m.groups()
      assert data is not None
      data = data.strip()
      if tool == "pytype" and data == "skip-file":
        # Abort immediately to avoid unnecessary processing.
        raise SkipFileError()
      if tool == "type" and open_ended and is_nested:
        # Discard type comments embedded in larger whole-line comments.
        continue
      structured_comment = _StructuredComment(line, tool, data, open_ended)
      for line_range, group in self._get_containing_groups(line):
        if not isinstance(line_range, Call):
          # A structured comment belongs to exactly one logical statement.
          group.append(structured_comment)
          break
        elif not open_ended and (
            tool == "pytype" or (tool == "type" and IGNORE_RE.match(data))):
          # A "type: ignore" or "pytype:" comment can additionally belong to any
          # number of overlapping function calls.
          group.append(structured_comment)
      else:
        raise AssertionError("Could not find a line range for comment "
                             f"{structured_comment} on line {line}")

  def _get_position(self, node):
    return self.get_metadata(libcst.metadata.PositionProvider, node)

  # Comments are found inside TrailingWhitespace and EmptyLine nodes. We visit
  # all the nodes that can contain a TrailingWhitespace node and add a comment
  # group for each of them, then populate the groups with TrailingWhitespace
  # comments. EmptyLine comments form their own single-line groups. Note that
  # comment.start should be used to get the line at which a comment is located;
  # comment.end is at column 0 of the following line.

  def _visit_comment_owner(self, node, cls=LineRange):
    pos = self._get_position(node)
    self._add_structured_comment_group(pos.start.line, pos.end.line, cls)

  def visit_Decorator(self, node):
    self._visit_comment_owner(node)

  def visit_SimpleStatementLine(self, node):
    self._visit_comment_owner(node)

  def visit_SimpleStatementSuite(self, node):
    self._visit_comment_owner(node)

  def visit_IndentedBlock(self, node):
    # An indented block has a "header" child that holds any trailing comment
    # from the block's header, e.g.:
    #   if __random__:  # header comment
    #     indented_block
    # The comment's line range goes from the first line of the header to the
    # comment line.
    parent = self.get_metadata(libcst.metadata.ParentNodeProvider, node)
    # visit_FunctionDef takes care of adding line ranges for FunctionDef.
    if not isinstance(parent, libcst.FunctionDef):
      start = self._get_position(parent).start
      end = self._get_position(node.header).start
      self._add_structured_comment_group(start.line, end.line)

  def visit_ParenthesizedWhitespace(self, node):
    self._visit_comment_owner(node)

  def visit_Call(self, node):
    self._visit_comment_owner(node, cls=Call)

  def visit_Comparison(self, node):
    self._visit_comment_owner(node, cls=Call)

  def visit_Subscript(self, node):
    self._visit_comment_owner(node, cls=Call)

  def visit_TrailingWhitespace(self, node):
    if node.comment:
      line = self._get_position(node).start.line
      self._process_comment(line, node.comment.value, open_ended=False)

  def visit_EmptyLine(self, node):
    if node.comment:
      line = self._get_position(node).start.line
      self._add_structured_comment_group(line, line)
      self._process_comment(line, node.comment.value, open_ended=True)

  def visit_AnnAssign(self, node):
    if not node.value:
      # vm.py preprocesses the source code so that all annotations in function
      # bodies have values. So the only annotations without values are module-
      # and class-level ones, which generate STORE opcodes and therefore
      # don't need to be handled here.
      return
    pos = self._get_position(node)
    # Gets a string representation of the annotation.
    annotation = re.sub(
        r"\s*(#.*)?\n\s*", "",
        libcst.Module([node.annotation.annotation]).code)
    self.variable_annotations.append(
        _VariableAnnotation(pos.start.line, pos.end.line, annotation))

  def visit_Return(self, node):
    self.returns.add(self._get_position(node).start.line)

  def _visit_decorators(self, node):
    if not node.decorators:
      return
    # The line range for this definition starts at the beginning of the last
    # decorator and ends at the definition's name.
    decorator = node.decorators[-1]
    start = self._get_position(decorator).start
    end = self._get_position(node.name).start
    self.decorators.append(LineRange(start.line, end.line))

  def _visit_def(self, node):
    line = self._get_position(node).start.line
    if not self.defs_start or line < self.defs_start:
      self.defs_start = line

  def visit_ClassDef(self, node):
    self._visit_decorators(node)
    self._visit_def(node)

  def visit_FunctionDef(self, node):
    # A function signature's line range starts at the beginning of the signature
    # and ends at the final colon.
    pos = self._get_position(node)
    self._add_structured_comment_group(
        pos.start.line,
        self._get_position(node.whitespace_before_colon).end.line)
    self._visit_decorators(node)
    self._visit_def(node)
    self.function_ranges[pos.start.line] = pos.end.line


def parse_src(src, python_version):
  """Parses a string of source code into a LibCST tree."""
  assert python_version < (3, 9)
  version_str = utils.format_version(python_version)
  config = libcst.PartialParserConfig(python_version=version_str)
  src_tree = libcst.parse_module(src, config)
  # The docstring for unsafe_skip_copy says:
  #
  # When true, this skips the deep cloning of the module.
  # This can provide a small performance benefit, but you should only use this
  # if you know that there are no duplicate nodes in your tree (e.g. this
  # module came from the parser).
  #
  # We only pass in trees constructed by libcst.parse_module, so we disable
  # copying for the performance benefit.
  return libcst.metadata.MetadataWrapper(src_tree, unsafe_skip_copy=True)


def visit_src_tree(src_tree):
  visitor = _ParseVisitor()
  try:
    src_tree.visit(visitor)
  except RecursionError:
    log.warning("File parsing failed. Comment directives and some variable "
                "annotations will be ignored.")
    return None
  return visitor



from typing import Any, TypeVar

from pygments.formatter import Formatter

_T = TypeVar("_T", str, bytes)

class NullFormatter(Formatter[_T]):
    name: str
    aliases: Any
    filenames: Any
    def format(self, tokensource, outfile) -> None: ...

class RawTokenFormatter(Formatter[_T]):
    name: str
    aliases: Any
    filenames: Any
    unicodeoutput: bool
    encoding: str
    compress: Any
    error_color: Any
    def format(self, tokensource, outfile) -> None: ...

class TestcaseFormatter(Formatter[_T]):
    name: str
    aliases: Any
    def format(self, tokensource, outfile) -> None: ...



from typing import Any

from selenium.common.exceptions import NoSuchElementException as NoSuchElementException, TimeoutException as TimeoutException

POLL_FREQUENCY: float
IGNORED_EXCEPTIONS: Any

class WebDriverWait:
    def __init__(self, driver, timeout, poll_frequency=..., ignored_exceptions: Any | None = ...) -> None: ...
    def until(self, method, message: str = ...): ...
    def until_not(self, method, message: str = ...): ...



import sys
import threading
from _typeshed import Self
from abc import abstractmethod
from collections.abc import Container, Iterable, Iterator, Sequence
from logging import Logger
from types import TracebackType
from typing import Any, Callable, Generic, Protocol, TypeVar, overload
from typing_extensions import Literal, ParamSpec, SupportsIndex

if sys.version_info >= (3, 9):
    from types import GenericAlias

FIRST_COMPLETED: Literal["FIRST_COMPLETED"]
FIRST_EXCEPTION: Literal["FIRST_EXCEPTION"]
ALL_COMPLETED: Literal["ALL_COMPLETED"]
PENDING: Literal["PENDING"]
RUNNING: Literal["RUNNING"]
CANCELLED: Literal["CANCELLED"]
CANCELLED_AND_NOTIFIED: Literal["CANCELLED_AND_NOTIFIED"]
FINISHED: Literal["FINISHED"]
_FUTURE_STATES: list[str]
_STATE_TO_DESCRIPTION_MAP: dict[str, str]
LOGGER: Logger

class Error(Exception): ...
class CancelledError(Error): ...
class TimeoutError(Error): ...

if sys.version_info >= (3, 8):
    class InvalidStateError(Error): ...

if sys.version_info >= (3, 7):
    class BrokenExecutor(RuntimeError): ...

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_P = ParamSpec("_P")

# Copied over Collection implementation as it does not exist in Python 2 and <3.6.
# Also to solve pytype issues with _Collection.
class _Collection(Iterable[_T_co], Container[_T_co], Protocol[_T_co]):
    # Implement Sized (but don't have it as a base class).
    @abstractmethod
    def __len__(self) -> int: ...

class Future(Generic[_T]):
    def __init__(self) -> None: ...
    def cancel(self) -> bool: ...
    def cancelled(self) -> bool: ...
    def running(self) -> bool: ...
    def done(self) -> bool: ...
    def add_done_callback(self, fn: Callable[[Future[_T]], Any]) -> None: ...
    def result(self, timeout: float | None = ...) -> _T: ...
    def set_running_or_notify_cancel(self) -> bool: ...
    def set_result(self, result: _T) -> None: ...
    def exception(self, timeout: float | None = ...) -> BaseException | None: ...
    def set_exception(self, exception: BaseException | None) -> None: ...
    if sys.version_info >= (3, 9):
        def __class_getitem__(cls, item: Any) -> GenericAlias: ...

class Executor:
    if sys.version_info >= (3, 9):
        def submit(self, __fn: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs) -> Future[_T]: ...
    else:
        def submit(self, fn: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs) -> Future[_T]: ...

    def map(
        self, fn: Callable[..., _T], *iterables: Iterable[Any], timeout: float | None = ..., chunksize: int = ...
    ) -> Iterator[_T]: ...
    if sys.version_info >= (3, 9):
        def shutdown(self, wait: bool = ..., *, cancel_futures: bool = ...) -> None: ...
    else:
        def shutdown(self, wait: bool = ...) -> None: ...

    def __enter__(self: Self) -> Self: ...
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> bool | None: ...

def as_completed(fs: Iterable[Future[_T]], timeout: float | None = ...) -> Iterator[Future[_T]]: ...

# Ideally this would be a namedtuple, but mypy doesn't support generic tuple types. See #1976
class DoneAndNotDoneFutures(Sequence[set[Future[_T]]]):
    if sys.version_info >= (3, 10):
        __match_args__ = ("done", "not_done")
    @property
    def done(self) -> set[Future[_T]]: ...
    @property
    def not_done(self) -> set[Future[_T]]: ...
    def __new__(_cls, done: set[Future[_T]], not_done: set[Future[_T]]) -> DoneAndNotDoneFutures[_T]: ...
    def __len__(self) -> int: ...
    @overload
    def __getitem__(self, __i: SupportsIndex) -> set[Future[_T]]: ...
    @overload
    def __getitem__(self, __s: slice) -> DoneAndNotDoneFutures[_T]: ...

def wait(fs: Iterable[Future[_T]], timeout: float | None = ..., return_when: str = ...) -> DoneAndNotDoneFutures[_T]: ...

class _Waiter:
    event: threading.Event
    finished_futures: list[Future[Any]]
    def __init__(self) -> None: ...
    def add_result(self, future: Future[Any]) -> None: ...
    def add_exception(self, future: Future[Any]) -> None: ...
    def add_cancelled(self, future: Future[Any]) -> None: ...

class _AsCompletedWaiter(_Waiter):
    lock: threading.Lock
    def __init__(self) -> None: ...
    def add_result(self, future: Future[Any]) -> None: ...
    def add_exception(self, future: Future[Any]) -> None: ...
    def add_cancelled(self, future: Future[Any]) -> None: ...

class _FirstCompletedWaiter(_Waiter):
    def add_result(self, future: Future[Any]) -> None: ...
    def add_exception(self, future: Future[Any]) -> None: ...
    def add_cancelled(self, future: Future[Any]) -> None: ...

class _AllCompletedWaiter(_Waiter):
    num_pending_calls: int
    stop_on_exception: bool
    lock: threading.Lock
    def __init__(self, num_pending_calls: int, stop_on_exception: bool) -> None: ...
    def add_result(self, future: Future[Any]) -> None: ...
    def add_exception(self, future: Future[Any]) -> None: ...
    def add_cancelled(self, future: Future[Any]) -> None: ...

class _AcquireFutures:
    futures: Iterable[Future[Any]]
    def __init__(self, futures: Iterable[Future[Any]]) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, *args: object) -> None: ...



from io import BufferedIOBase, BufferedRandom, BytesIO
from typing import Any, Callable

COPY_BYTES: int
STRBUF_LIMIT: int

class FileBasedBuffer:
    remain: int = ...
    file: BytesIO = ...
    def __init__(self, file: BytesIO, from_buffer: BytesIO | None = ...) -> None: ...
    def __len__(self) -> int: ...
    def __nonzero__(self) -> bool: ...
    __bool__: Callable[[], bool] = ...
    def append(self, s: Any) -> None: ...
    def get(self, numbytes: int = ..., skip: bool = ...) -> bytes: ...
    def skip(self, numbytes: int, allow_prune: int = ...) -> None: ...
    def newfile(self) -> Any: ...
    def prune(self) -> None: ...
    def getfile(self) -> Any: ...
    def close(self) -> None: ...

class TempfileBasedBuffer(FileBasedBuffer):
    def __init__(self, from_buffer: BytesIO | None = ...) -> None: ...
    def newfile(self) -> BufferedRandom: ...

class BytesIOBasedBuffer(FileBasedBuffer):
    file: BytesIO = ...
    def __init__(self, from_buffer: BytesIO | None = ...) -> None: ...
    def newfile(self) -> BytesIO: ...

class ReadOnlyFileBasedBuffer(FileBasedBuffer):
    file: BytesIO = ...
    block_size: int = ...
    def __init__(self, file: BytesIO, block_size: int = ...) -> None: ...
    remain: int = ...
    def prepare(self, size: int | None = ...) -> int: ...
    def get(self, numbytes: int = ..., skip: bool = ...) -> bytes: ...
    def __iter__(self) -> ReadOnlyFileBasedBuffer: ...
    def next(self) -> bytes | None: ...
    __next__: Callable[[], bytes | None] = ...
    def append(self, s: Any) -> None: ...

class OverflowableBuffer:
    overflowed: bool = ...
    buf: BufferedIOBase | None = ...
    strbuf: bytes = ...
    overflow: int = ...
    def __init__(self, overflow: int) -> None: ...
    def __len__(self) -> int: ...
    def __nonzero__(self) -> bool: ...
    __bool__: Callable[[], bool] = ...
    def append(self, s: bytes) -> None: ...
    def get(self, numbytes: int = ..., skip: bool = ...) -> bytes: ...
    def skip(self, numbytes: int, allow_prune: bool = ...) -> None: ...
    def prune(self) -> None: ...
    def getfile(self) -> BytesIO: ...
    def close(self) -> None: ...



import datetime
from _typeshed import Self
from typing import Iterable

from dateutil.relativedelta import relativedelta

class DateTimeRange:
    NOT_A_TIME_STR: str
    start_time_format: str
    end_time_format: str
    is_output_elapse: bool
    separator: str
    def __init__(
        self,
        start_datetime: datetime.datetime | str | None = ...,
        end_datetime: datetime.datetime | str | None = ...,
        start_time_format: str = ...,
        end_time_format: str = ...,
    ) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __add__(self, other: datetime.timedelta) -> DateTimeRange: ...
    def __iadd__(self: Self, other: datetime.timedelta) -> Self: ...
    def __sub__(self, other: datetime.timedelta) -> DateTimeRange: ...
    def __isub__(self: Self, other: datetime.timedelta) -> Self: ...
    def __contains__(self, x: datetime.timedelta | datetime.datetime | DateTimeRange | str) -> bool: ...
    @property
    def start_datetime(self) -> datetime.datetime: ...
    @property
    def end_datetime(self) -> datetime.datetime: ...
    @property
    def timedelta(self) -> datetime.timedelta: ...
    def is_set(self) -> bool: ...
    def validate_time_inversion(self) -> None: ...
    def is_valid_timerange(self) -> bool: ...
    def is_intersection(self, x: DateTimeRange) -> bool: ...
    def get_start_time_str(self) -> str: ...
    def get_end_time_str(self) -> str: ...
    def get_timedelta_second(self) -> float: ...
    def set_start_datetime(self, value: datetime.datetime | str | None, timezone: str | None = ...) -> None: ...
    def set_end_datetime(self, value: datetime.datetime | str | None, timezone: str | None = ...) -> None: ...
    def set_time_range(self, start: datetime.datetime | str | None, end: datetime.datetime | str | None) -> None: ...
    def range(self, step: datetime.timedelta | relativedelta) -> Iterable[datetime.datetime]: ...
    def intersection(self, x: DateTimeRange) -> DateTimeRange: ...
    def encompass(self, x: DateTimeRange) -> DateTimeRange: ...
    def truncate(self, percentage: float) -> None: ...



from .base import (
    get_canonical_version as get_canonical_version,
    in_stdlib as in_stdlib,
    long_versions as long_versions,
    short_versions as short_versions,
    stdlib_list as stdlib_list,
)

__version__: str



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class Extension(Serialisable):
    tagname: str
    uri: Any
    def __init__(self, uri: Any | None = ...) -> None: ...

class ExtensionList(Serialisable):
    tagname: str
    ext: Any
    __elements__: Any
    def __init__(self, ext=...) -> None: ...

class IgnoredError(Serialisable):
    tagname: str
    sqref: Any
    evalError: Any
    twoDigitTextYear: Any
    numberStoredAsText: Any
    formula: Any
    formulaRange: Any
    unlockedFormula: Any
    emptyCellReference: Any
    listDataValidation: Any
    calculatedColumn: Any
    def __init__(
        self,
        sqref: Any | None = ...,
        evalError: bool = ...,
        twoDigitTextYear: bool = ...,
        numberStoredAsText: bool = ...,
        formula: bool = ...,
        formulaRange: bool = ...,
        unlockedFormula: bool = ...,
        emptyCellReference: bool = ...,
        listDataValidation: bool = ...,
        calculatedColumn: bool = ...,
    ) -> None: ...

class IgnoredErrors(Serialisable):
    tagname: str
    ignoredError: Any
    extLst: Any
    __elements__: Any
    def __init__(self, ignoredError=..., extLst: Any | None = ...) -> None: ...



import sys

if sys.platform == "win32":
    from distutils.cmd import Command

    class bdist_msi(Command):
        def initialize_options(self) -> None: ...
        def finalize_options(self) -> None: ...
        def run(self) -> None: ...



from typing import Any

LC_CTYPE: Any
PLURALS: Any
DEFAULT_PLURAL: Any

class _PluralTuple(tuple[int, str]):
    @property
    def num_plurals(self) -> int: ...
    @property
    def plural_expr(self) -> str: ...
    @property
    def plural_forms(self) -> str: ...

def get_plural(locale=...): ...



from typing import Any

class Composable:
    def __init__(self, wrapped) -> None: ...
    def as_string(self, context) -> None: ...
    def __add__(self, other): ...
    def __mul__(self, n): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...

class Composed(Composable):
    def __init__(self, seq) -> None: ...
    @property
    def seq(self): ...
    def as_string(self, context): ...
    def __iter__(self): ...
    def __add__(self, other): ...
    def join(self, joiner): ...

class SQL(Composable):
    def __init__(self, string) -> None: ...
    @property
    def string(self): ...
    def as_string(self, context): ...
    def format(self, *args, **kwargs): ...
    def join(self, seq): ...

class Identifier(Composable):
    def __init__(self, *strings) -> None: ...
    @property
    def strings(self): ...
    @property
    def string(self): ...
    def as_string(self, context): ...

class Literal(Composable):
    @property
    def wrapped(self): ...
    def as_string(self, context): ...

class Placeholder(Composable):
    def __init__(self, name: Any | None = ...) -> None: ...
    @property
    def name(self): ...
    def as_string(self, context): ...

NULL: Any
DEFAULT: Any
